#include "Win32WebServer.h"

#include "Win32TempFile.h"
#include "Win32Buffer.h"
#include "Win32WebResponse.h"
#include "Win32WebRequest.h"

#include "AbstractWebApp.h"
#include "HttpRequest.h"
#include "HttpResponse.h"
#include "FileLogger.h"
#include "StringUtils.h"

Win32WebServer::Win32WebServer(AbstractWebApp* webApp)
	: mListenUrl("http://localhost:80/49152"),
        mWebApp(webApp)
{

}

Win32WebServer::~Win32WebServer()
{
	clear();
}

void Win32WebServer::clear()
{
	// Remove url
	::HttpRemoveUrl(mWorkingQueue, StringUtils::convert(mListenUrl).c_str());

	// Close the Request Queue handle.
	if (mWorkingQueue)
	{
		::CloseHandle(mWorkingQueue);
	}

	// Call HttpTerminate.
	::HttpTerminate(HTTP_INITIALIZE_SERVER, nullptr);
}

void Win32WebServer::initialize()
{
	HTTPAPI_VERSION HttpApiVersion = HTTPAPI_VERSION_1;

    auto retCode = ::HttpInitialize(HttpApiVersion, HTTP_INITIALIZE_SERVER, nullptr);
	if (FAILED(retCode))
	{
		MLOG_ERROR("Failed to initialize win32 http server.");
		return;
	}

	retCode = ::HttpCreateHttpHandle(&mWorkingQueue, 0);
	if (FAILED(retCode))
	{
		MLOG_ERROR("Failed to create request queue handle.");
		return;
	}

	retCode = ::HttpAddUrl(mWorkingQueue, StringUtils::convert(mListenUrl).c_str(), nullptr);
	if (FAILED(retCode))
	{
		MLOG_ERROR("Failed to register queue to url: " << mListenUrl);
		return;
	}
}

DWORD Win32WebServer::onRequest(const Win32WebRequest& request)
{
    auto app_request = request.getRequest();
    auto response = mWebApp->onHttpRequest(app_request);

    DWORD result;
    if (app_request.getVerb() == HttpRequest::Verb::POST)
    {
        result = sendHttpPostResponse(request.getHandle());
    }
    else
    {
        result = sendHttpResponse(request.getHandle(), response);
    }

    /*
    case HttpVerbGET:
        MLOG_INFO("Got a GET request for: " << url);
        result = sendHttpResponse(request, 200, "OK", "Hey! You hit the server \r\n");
        break;
    case HttpVerbPOST:
        MLOG_INFO("Got a POST request for: " << url);
        result = sendHttpPostResponse(request);
        break;
    default:
        MLOG_INFO("Got an unknown request for: " << url);
        result = sendHttpResponse(request, 503, "Not Implemented");
        break;
    */

    return result;
}

DWORD Win32WebServer::sendHttpResponse(PHTTP_REQUEST pRequest, const HttpResponse& appResponse)
{
    Win32WebResponse response(appResponse);

    DWORD bytesSent;
    const auto result = ::HttpSendHttpResponse(mWorkingQueue, pRequest->RequestId, 0, &(response.getResponse()), NULL, &bytesSent, NULL, 0, NULL, NULL);
    if (result != NO_ERROR)
    {
        MLOG_ERROR("Http response failed with error: " << result);
    }
    return result;
}

#define MAX_ULONG_STR ((ULONG) sizeof("4294967295"))

DWORD Win32WebServer::sendHttpPostResponse(PHTTP_REQUEST pRequest)
{
    // Allocate space for an entity buffer. Buffer can be increased on demand.
    const ULONG entity_buffer_length = 2048;

    Win32Buffer entity_buffer(entity_buffer_length);
    if (!entity_buffer.isValid())
    {
        return ERROR_NOT_ENOUGH_MEMORY;
    }

    HTTP_RESPONSE response;
    initializeHttpResponse(response, 200, "OK");

    DWORD result{ 0 };
    if (pRequest->Flags & HTTP_REQUEST_FLAG_MORE_ENTITY_BODY_EXISTS)
    {
        Win32TempFile temp_file;
        if (!temp_file.isValid())
        {
            return 0;
        }

        ULONG total_bytes_read = 0;
        ULONG bytes_read = 0;      
        do
        {
            // Read the entity chunk from the request.
            bytes_read = 0;
            result = ::HttpReceiveRequestEntityBody(mWorkingQueue, pRequest->RequestId, 0, entity_buffer.getBuffer(), entity_buffer_length, &bytes_read, nullptr);

            switch (result)
            {
            case NO_ERROR:
                if (bytes_read != 0)
                {
                    total_bytes_read += bytes_read;
                    temp_file.write(entity_buffer.getBufferAsUChar(), bytes_read);
                }
                break;
            case ERROR_HANDLE_EOF:
                // The last request entity body has been read. Send back a response. 
                //
                // To illustrate entity sends via HttpSendResponseEntityBody, the response will 
                // be sent over multiple calls. To do this, pass the HTTP_SEND_RESPONSE_FLAG_MORE_DATA flag.

                if (bytes_read != 0)
                {
                    total_bytes_read += bytes_read;
                    temp_file.write(entity_buffer.getBufferAsUChar(), bytes_read);
                }

                // Because the response is sent over multiple API calls, add a content-length.
                // Alternatively, the response could have been sent using chunked transfer encoding, by passimg "Transfer-Encoding: Chunked".
                // NOTE: Because the TotalBytesread in a ULONG are accumulated, this will not work for entity bodies larger than 4 GB. 
                // For support of large entity bodies, use a ULONGLONG.
                CHAR szContentLength[MAX_ULONG_STR];
                sprintf_s(szContentLength, MAX_ULONG_STR, "%lu", total_bytes_read);
                response.Headers.KnownHeaders[HttpHeaderContentLength].pRawValue = szContentLength;
                response.Headers.KnownHeaders[HttpHeaderContentLength].RawValueLength = (USHORT)strlen(szContentLength);

                DWORD bytesSent;
                result = ::HttpSendHttpResponse(mWorkingQueue, pRequest->RequestId, HTTP_SEND_RESPONSE_FLAG_MORE_DATA, &response, NULL, &bytesSent, NULL, 0, NULL, NULL);
                if (result != NO_ERROR)
                {
                    MLOG_ERROR("HttpSendHttpResponse failed with: " << result);
                    return;
                }

                // Send entity body from a file handle.
                HTTP_DATA_CHUNK dataChunk;
                dataChunk.DataChunkType = HttpDataChunkFromFileHandle;
                dataChunk.FromFileHandle.ByteRange.StartingOffset.QuadPart = 0;
                dataChunk.FromFileHandle.ByteRange.Length.QuadPart = HTTP_BYTE_RANGE_TO_EOF;
                dataChunk.FromFileHandle.FileHandle = temp_file.getHandle();

                result = ::HttpSendResponseEntityBody(mWorkingQueue, pRequest->RequestId, 0, 1, &dataChunk, NULL, NULL, 0, NULL, NULL);
                if (result != NO_ERROR)
                {
                    MLOG_ERROR("HttpSendResponseEntityBody failed" << result);
                }
                break;
            default:
                MLOG_ERROR("HttpReceiveRequestEntityBody failed with: " << result);
                break;
            }

        } while (true);
    }
    else
    {
        DWORD bytesSent;
        result = ::HttpSendHttpResponse(mWorkingQueue, pRequest->RequestId, 0, &response, NULL, &bytesSent, NULL, 0, NULL, NULL );
        if (result != NO_ERROR)
        {
            MLOG_ERROR("HttpSendHttpResponse failed with: " << result);
        }
    }
    return result;
}

void Win32WebServer::run()
{
    Win32WebRequest request;
    if (!request.isValid())
    {
        MLOG_ERROR("Failed to allocate http request buffer");
        return;
    }

    HTTP_REQUEST_ID request_id;
    HTTP_SET_NULL_ID(&request_id);
    while(true)
    {
        request.zeroBufferMemory();

        DWORD bytes_read;
        auto result = ::HttpReceiveHttpRequest(mWorkingQueue, request_id, 0, request.getHandle(), request.getBufferSize(), &bytes_read, nullptr);
        if (NO_ERROR == result)
        {
            result = onRequest(request);
            if (result != NO_ERROR)
            {
                break;
            }
            HTTP_SET_NULL_ID(&request_id);
        }
        else if (result == ERROR_MORE_DATA)
        {
            // The input buffer was too small to hold the request headers. Increase the buffer size and call the API again. 
            request_id = request.getHandle()->RequestId;

            request.resizeBuffer(bytes_read);
            if (!request.isValid())
            {
                result = ERROR_NOT_ENOUGH_MEMORY;
                break;
            }
        }
        else if (ERROR_CONNECTION_INVALID == result && !HTTP_IS_NULL_ID(&request_id))
        {
            // The TCP connection was corrupted by the peer when attempting to handle a request with more buffer. Continue to the next request.
            HTTP_SET_NULL_ID(&request_id);
        }
        else
        {
            break;
        }
    }
}